# Copyright: See the LICENSE file.

import datetime
import unittest
from unittest import mock

from factory import base, declarations, errors, helpers

from . import utils


class OrderedDeclarationTestCase(unittest.TestCase):
    def test_errors(self):
        with self.assertRaises(NotImplementedError):
            utils.evaluate_declaration(declarations.OrderedDeclaration())


class DigTestCase(unittest.TestCase):
    class MyObj:
        def __init__(self, n):
            self.n = n

    def test_chaining(self):
        obj = self.MyObj(1)
        obj.a = self.MyObj(2)
        obj.a.b = self.MyObj(3)
        obj.a.b.c = self.MyObj(4)

        self.assertEqual(2, declarations.deepgetattr(obj, 'a').n)
        with self.assertRaises(AttributeError):
            declarations.deepgetattr(obj, 'b')
        self.assertEqual(2, declarations.deepgetattr(obj, 'a.n'))
        self.assertEqual(3, declarations.deepgetattr(obj, 'a.c', 3))
        with self.assertRaises(AttributeError):
            declarations.deepgetattr(obj, 'a.c.n')
        with self.assertRaises(AttributeError):
            declarations.deepgetattr(obj, 'a.d')
        self.assertEqual(3, declarations.deepgetattr(obj, 'a.b').n)
        self.assertEqual(3, declarations.deepgetattr(obj, 'a.b.n'))
        self.assertEqual(4, declarations.deepgetattr(obj, 'a.b.c').n)
        self.assertEqual(4, declarations.deepgetattr(obj, 'a.b.c.n'))
        self.assertEqual(42, declarations.deepgetattr(obj, 'a.b.c.n.x', 42))


class MaybeTestCase(unittest.TestCase):
    def test_init(self):
        declarations.Maybe('foo', 1, 2)

        with self.assertRaisesRegex(TypeError, 'Inconsistent phases'):
            declarations.Maybe('foo', declarations.LazyAttribute(None), declarations.PostGenerationDeclaration())


class SelfAttributeTestCase(unittest.TestCase):
    def test_standard(self):
        a = declarations.SelfAttribute('foo.bar.baz')
        self.assertEqual(0, a.depth)
        self.assertEqual('foo.bar.baz', a.attribute_name)
        self.assertEqual(declarations._UNSPECIFIED, a.default)

    def test_dot(self):
        a = declarations.SelfAttribute('.bar.baz')
        self.assertEqual(1, a.depth)
        self.assertEqual('bar.baz', a.attribute_name)
        self.assertEqual(declarations._UNSPECIFIED, a.default)

    def test_default(self):
        a = declarations.SelfAttribute('bar.baz', 42)
        self.assertEqual(0, a.depth)
        self.assertEqual('bar.baz', a.attribute_name)
        self.assertEqual(42, a.default)

    def test_parent(self):
        a = declarations.SelfAttribute('..bar.baz')
        self.assertEqual(2, a.depth)
        self.assertEqual('bar.baz', a.attribute_name)
        self.assertEqual(declarations._UNSPECIFIED, a.default)

    def test_grandparent(self):
        a = declarations.SelfAttribute('...bar.baz')
        self.assertEqual(3, a.depth)
        self.assertEqual('bar.baz', a.attribute_name)
        self.assertEqual(declarations._UNSPECIFIED, a.default)


class IteratorTestCase(unittest.TestCase):
    def test_cycle(self):
        it = declarations.Iterator([1, 2])
        self.assertEqual(1, utils.evaluate_declaration(it, force_sequence=0))
        self.assertEqual(2, utils.evaluate_declaration(it, force_sequence=1))
        self.assertEqual(1, utils.evaluate_declaration(it, force_sequence=2))
        self.assertEqual(2, utils.evaluate_declaration(it, force_sequence=3))

    def test_no_cycling(self):
        it = declarations.Iterator([1, 2], cycle=False)
        self.assertEqual(1, utils.evaluate_declaration(it, force_sequence=0))
        self.assertEqual(2, utils.evaluate_declaration(it, force_sequence=1))
        with self.assertRaises(StopIteration):
            utils.evaluate_declaration(it, force_sequence=2)

    def test_initial_reset(self):
        it = declarations.Iterator([1, 2])
        it.reset()

    def test_reset_cycle(self):
        it = declarations.Iterator([1, 2])
        self.assertEqual(1, utils.evaluate_declaration(it, force_sequence=0))
        self.assertEqual(2, utils.evaluate_declaration(it, force_sequence=1))
        self.assertEqual(1, utils.evaluate_declaration(it, force_sequence=2))
        self.assertEqual(2, utils.evaluate_declaration(it, force_sequence=3))
        self.assertEqual(1, utils.evaluate_declaration(it, force_sequence=4))
        it.reset()
        self.assertEqual(1, utils.evaluate_declaration(it, force_sequence=5))
        self.assertEqual(2, utils.evaluate_declaration(it, force_sequence=6))

    def test_reset_no_cycling(self):
        it = declarations.Iterator([1, 2], cycle=False)
        self.assertEqual(1, utils.evaluate_declaration(it, force_sequence=0))
        self.assertEqual(2, utils.evaluate_declaration(it, force_sequence=1))
        with self.assertRaises(StopIteration):
            utils.evaluate_declaration(it, force_sequence=2)
        it.reset()
        self.assertEqual(1, utils.evaluate_declaration(it, force_sequence=0))
        self.assertEqual(2, utils.evaluate_declaration(it, force_sequence=1))
        with self.assertRaises(StopIteration):
            utils.evaluate_declaration(it, force_sequence=2)

    def test_getter(self):
        it = declarations.Iterator([(1, 2), (1, 3)], getter=lambda p: p[1])
        self.assertEqual(2, utils.evaluate_declaration(it, force_sequence=0))
        self.assertEqual(3, utils.evaluate_declaration(it, force_sequence=1))
        self.assertEqual(2, utils.evaluate_declaration(it, force_sequence=2))
        self.assertEqual(3, utils.evaluate_declaration(it, force_sequence=3))


class TransformerTestCase(unittest.TestCase):
    def test_transform(self):
        t = declarations.Transformer('foo', transform=str.upper)
        self.assertEqual("FOO", utils.evaluate_declaration(t))


class PostGenerationDeclarationTestCase(unittest.TestCase):
    def test_post_generation(self):
        call_params = []

        def foo(*args, **kwargs):
            call_params.append(args)
            call_params.append(kwargs)

        helpers.build(
            dict,
            foo=declarations.PostGeneration(foo),
            foo__bar=42,
            blah=42,
            blah__baz=1,
        )

        self.assertEqual(2, len(call_params))
        self.assertEqual(3, len(call_params[0]))  # instance, step, context.value
        self.assertEqual({'bar': 42}, call_params[1])

    def test_decorator_simple(self):
        call_params = []

        @helpers.post_generation
        def foo(*args, **kwargs):
            call_params.append(args)
            call_params.append(kwargs)

        helpers.build(
            dict,
            foo=foo,
            foo__bar=42,
            blah=42,
            blah__baz=1,
        )

        self.assertEqual(2, len(call_params))
        self.assertEqual(3, len(call_params[0]))  # instance, step, context.value
        self.assertEqual({'bar': 42}, call_params[1])


class FactoryWrapperTestCase(unittest.TestCase):
    def test_invalid_path(self):
        with self.assertRaises(ValueError):
            declarations._FactoryWrapper('UnqualifiedSymbol')
        with self.assertRaises(ValueError):
            declarations._FactoryWrapper(42)

    def test_class(self):
        w = declarations._FactoryWrapper(datetime.date)
        self.assertEqual(datetime.date, w.get())

    def test_path(self):
        w = declarations._FactoryWrapper('datetime.date')
        self.assertEqual(datetime.date, w.get())

    def test_lazyness(self):
        f = declarations._FactoryWrapper('factory.declarations.Sequence')
        self.assertEqual(None, f.factory)

        factory_class = f.get()
        self.assertEqual(declarations.Sequence, factory_class)

    def test_cache(self):
        """Ensure that _FactoryWrapper tries to import only once."""
        orig_date = datetime.date
        w = declarations._FactoryWrapper('datetime.date')
        self.assertEqual(None, w.factory)

        factory_class = w.get()
        self.assertEqual(orig_date, factory_class)

        try:
            # Modify original value
            datetime.date = None
            # Repeat import
            factory_class = w.get()
            self.assertEqual(orig_date, factory_class)

        finally:
            # IMPORTANT: restore attribute.
            datetime.date = orig_date


class PostGenerationMethodCallTestCase(unittest.TestCase):
    def build(self, declaration, **params):
        f = helpers.make_factory(mock.MagicMock, post=declaration)
        return f(**params)

    def test_simplest_setup_and_call(self):
        obj = self.build(
            declarations.PostGenerationMethodCall('method'),
        )
        obj.method.assert_called_once_with()

    def test_call_with_method_args(self):
        obj = self.build(
            declarations.PostGenerationMethodCall('method', 'data'),
        )
        obj.method.assert_called_once_with('data')

    def test_call_with_passed_extracted_string(self):
        obj = self.build(
            declarations.PostGenerationMethodCall('method'),
            post='data',
        )
        obj.method.assert_called_once_with('data')

    def test_call_with_passed_extracted_int(self):
        obj = self.build(
            declarations.PostGenerationMethodCall('method'),
            post=1,
        )
        obj.method.assert_called_once_with(1)

    def test_call_with_passed_extracted_iterable(self):
        obj = self.build(
            declarations.PostGenerationMethodCall('method'),
            post=(1, 2, 3),
        )
        obj.method.assert_called_once_with((1, 2, 3))

    def test_call_with_method_kwargs(self):
        obj = self.build(
            declarations.PostGenerationMethodCall('method', data='data'),
        )
        obj.method.assert_called_once_with(data='data')

    def test_call_with_passed_kwargs(self):
        obj = self.build(
            declarations.PostGenerationMethodCall('method'),
            post__data='other',
        )
        obj.method.assert_called_once_with(data='other')

    def test_multi_call_with_multi_method_args(self):
        with self.assertRaises(errors.InvalidDeclarationError):
            self.build(
                declarations.PostGenerationMethodCall('method', 'arg1', 'arg2'),
            )


class PostGenerationOrdering(unittest.TestCase):

    def test_post_generation_declaration_order(self):
        postgen_results = []

        class Related(base.Factory):
            class Meta:
                model = mock.MagicMock()

        class Ordered(base.Factory):
            class Meta:
                model = mock.MagicMock()

            a = declarations.RelatedFactory(Related)
            z = declarations.RelatedFactory(Related)

            @helpers.post_generation
            def a1(*args, **kwargs):
                postgen_results.append('a1')

            @helpers.post_generation
            def zz(*args, **kwargs):
                postgen_results.append('zz')

            @helpers.post_generation
            def aa(*args, **kwargs):
                postgen_results.append('aa')

        postgen_names = Ordered._meta.post_declarations.sorted()
        self.assertEqual(postgen_names, ['a', 'z', 'a1', 'zz', 'aa'])

        # Test generation happens in desired order
        Ordered()
        self.assertEqual(postgen_results, ['a1', 'zz', 'aa'])



import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='TaggedItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('tag', models.SlugField()),
                ('object_id', models.PositiveIntegerField()),
                (
                    'content_type',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to='contenttypes.ContentType'
                    )
                ),
            ],
        ),
    ]



# Copyright: See the LICENSE file.

from flask import Flask
from flask_sqlalchemy import SQLAlchemy

app = Flask(__name__)
app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:////tmp/test.db'
db = SQLAlchemy(app)


class User(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String(80), unique=True)
    email = db.Column(db.String(120), unique=True)

    def __init__(self, username, email):
        self.username = username
        self.email = email

    def __repr__(self):
        return '<User %r>' % self.username


class UserLog(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    message = db.Column(db.String(1000))

    user_id = db.Column(db.Integer, db.ForeignKey('user.id'))
    user = db.relationship('User', backref=db.backref('logs', lazy='dynamic'))

    def __init__(self, message, user):
        self.message = message
        self.user = user

    def __repr__(self):
        return f'<Log for {self.user!r}: {self.message}>'



# Copyright: See the LICENSE file.

"""Settings for factory_boy/Django tests."""

import os

FACTORY_ROOT = os.path.join(
    os.path.abspath(os.path.dirname(__file__)),     # /path/to/fboy/tests/djapp/
    os.pardir,                                      # /path/to/fboy/tests/
    os.pardir,                                      # /path/to/fboy
)

MEDIA_ROOT = os.path.join(FACTORY_ROOT, 'tmp_test')

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
    },
    'replica': {
        'ENGINE': 'django.db.backends.sqlite3',
    },
}


INSTALLED_APPS = [
    'tests.djapp'
]

MIDDLEWARE_CLASSES = ()

SECRET_KEY = 'testing.'

# TODO: Will be the default after Django 5.0. Remove this setting when
# Django 5.0 is the last supported version.
USE_TZ = True



# Copyright: See the LICENSE file.

"""Tests for factory_boy/SQLAlchemy interactions."""

import unittest
from unittest import mock

try:
    import sqlalchemy
except ImportError:
    raise unittest.SkipTest("sqlalchemy tests disabled.")

import factory
from factory.alchemy import SQLAlchemyModelFactory

from .alchemyapp import models


class StandardFactory(SQLAlchemyModelFactory):
    class Meta:
        model = models.StandardModel
        sqlalchemy_session = models.session

    id = factory.Sequence(lambda n: n)
    foo = factory.Sequence(lambda n: 'foo%d' % n)


class NonIntegerPkFactory(SQLAlchemyModelFactory):
    class Meta:
        model = models.NonIntegerPk
        sqlalchemy_session = models.session

    id = factory.Sequence(lambda n: 'foo%d' % n)


class NoSessionFactory(SQLAlchemyModelFactory):
    class Meta:
        model = models.StandardModel
        sqlalchemy_session = None

    id = factory.Sequence(lambda n: n)


class MultifieldModelFactory(SQLAlchemyModelFactory):
    class Meta:
        model = models.MultiFieldModel
        sqlalchemy_get_or_create = ('slug',)
        sqlalchemy_session = models.session
        sqlalchemy_session_persistence = 'commit'

    id = factory.Sequence(lambda n: n)
    foo = factory.Sequence(lambda n: 'foo%d' % n)


class WithGetOrCreateFieldFactory(SQLAlchemyModelFactory):
    class Meta:
        model = models.StandardModel
        sqlalchemy_get_or_create = ('foo',)
        sqlalchemy_session = models.session
        sqlalchemy_session_persistence = 'commit'

    id = factory.Sequence(lambda n: n)
    foo = factory.Sequence(lambda n: 'foo%d' % n)


class WithMultipleGetOrCreateFieldsFactory(SQLAlchemyModelFactory):
    class Meta:
        model = models.MultifieldUniqueModel
        sqlalchemy_get_or_create = ("slug", "text",)
        sqlalchemy_session = models.session
        sqlalchemy_session_persistence = 'commit'

    id = factory.Sequence(lambda n: n)
    slug = factory.Sequence(lambda n: "slug%s" % n)
    text = factory.Sequence(lambda n: "text%s" % n)


class TransactionTestCase(unittest.TestCase):
    def setUp(self):
        models.Base.metadata.create_all(models.engine)

    def tearDown(self):
        models.session.remove()
        models.Base.metadata.drop_all(models.engine)


class SQLAlchemyPkSequenceTestCase(TransactionTestCase):
    def setUp(self):
        super().setUp()
        StandardFactory.reset_sequence(1)

    def test_pk_first(self):
        std = StandardFactory.build()
        self.assertEqual('foo1', std.foo)

    def test_pk_many(self):
        std1 = StandardFactory.build()
        std2 = StandardFactory.build()
        self.assertEqual('foo1', std1.foo)
        self.assertEqual('foo2', std2.foo)

    def test_pk_creation(self):
        std1 = StandardFactory.create()
        self.assertEqual('foo1', std1.foo)
        self.assertEqual(1, std1.id)

        StandardFactory.reset_sequence()
        std2 = StandardFactory.create()
        self.assertEqual('foo0', std2.foo)
        self.assertEqual(0, std2.id)

    def test_pk_force_value(self):
        std1 = StandardFactory.create(id=10)
        self.assertEqual('foo1', std1.foo)  # sequence and pk are unrelated
        self.assertEqual(10, std1.id)

        StandardFactory.reset_sequence()
        std2 = StandardFactory.create()
        self.assertEqual('foo0', std2.foo)  # Sequence doesn't care about pk
        self.assertEqual(0, std2.id)


class SQLAlchemyGetOrCreateTests(TransactionTestCase):
    def test_simple_call(self):
        obj1 = WithGetOrCreateFieldFactory(foo='foo1')
        obj2 = WithGetOrCreateFieldFactory(foo='foo1')
        self.assertEqual(obj1, obj2)

    def test_missing_arg(self):
        with self.assertRaises(factory.FactoryError):
            MultifieldModelFactory()

    def test_raises_exception_when_existing_objs(self):
        StandardFactory.create_batch(2, foo='foo')
        with self.assertRaises(sqlalchemy.orm.exc.MultipleResultsFound):
            WithGetOrCreateFieldFactory(foo='foo')

    def test_multicall(self):
        objs = MultifieldModelFactory.create_batch(
            6,
            slug=factory.Iterator(['main', 'alt']),
        )
        self.assertEqual(6, len(objs))
        self.assertEqual(2, len(set(objs)))
        self.assertEqual(
            list(
                obj.slug for obj in models.session.query(
                    models.MultiFieldModel.slug
                ).order_by(models.MultiFieldModel.slug)
            ),
            ["alt", "main"],
        )


class MultipleGetOrCreateFieldsTest(TransactionTestCase):
    def test_one_defined(self):
        obj1 = WithMultipleGetOrCreateFieldsFactory()
        obj2 = WithMultipleGetOrCreateFieldsFactory(slug=obj1.slug)
        self.assertEqual(obj1, obj2)

    def test_both_defined(self):
        obj1 = WithMultipleGetOrCreateFieldsFactory()
        with self.assertRaises(sqlalchemy.exc.IntegrityError):
            WithMultipleGetOrCreateFieldsFactory(slug=obj1.slug, text="alt")

    def test_unique_field_not_in_get_or_create(self):
        WithMultipleGetOrCreateFieldsFactory(title='Title')
        with self.assertRaises(sqlalchemy.exc.IntegrityError):
            WithMultipleGetOrCreateFieldsFactory(title='Title')


class SQLAlchemySessionPersistenceTestCase(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.mock_session = mock.NonCallableMagicMock(spec=models.session)

    def test_flushing(self):
        class FlushingPersistenceFactory(StandardFactory):
            class Meta:
                sqlalchemy_session = self.mock_session
                sqlalchemy_session_persistence = 'flush'

        self.mock_session.commit.assert_not_called()
        self.mock_session.flush.assert_not_called()

        FlushingPersistenceFactory.create()
        self.mock_session.commit.assert_not_called()
        self.mock_session.flush.assert_called_once_with()

    def test_committing(self):
        class CommittingPersistenceFactory(StandardFactory):
            class Meta:
                sqlalchemy_session = self.mock_session
                sqlalchemy_session_persistence = 'commit'

        self.mock_session.commit.assert_not_called()
        self.mock_session.flush.assert_not_called()

        CommittingPersistenceFactory.create()
        self.mock_session.commit.assert_called_once_with()
        self.mock_session.flush.assert_not_called()

    def test_noflush_nocommit(self):
        class InactivePersistenceFactory(StandardFactory):
            class Meta:
                sqlalchemy_session = self.mock_session
                sqlalchemy_session_persistence = None

        self.mock_session.commit.assert_not_called()
        self.mock_session.flush.assert_not_called()

        InactivePersistenceFactory.create()
        self.mock_session.commit.assert_not_called()
        self.mock_session.flush.assert_not_called()

    def test_type_error(self):
        with self.assertRaises(TypeError):
            class BadPersistenceFactory(StandardFactory):
                class Meta:
                    sqlalchemy_session_persistence = 'invalid_persistence_option'
                    model = models.StandardModel


class SQLAlchemyNonIntegerPkTestCase(TransactionTestCase):
    def tearDown(self):
        super().tearDown()
        NonIntegerPkFactory.reset_sequence()

    def test_first(self):
        nonint = NonIntegerPkFactory.build()
        self.assertEqual('foo0', nonint.id)

    def test_many(self):
        nonint1 = NonIntegerPkFactory.build()
        nonint2 = NonIntegerPkFactory.build()

        self.assertEqual('foo0', nonint1.id)
        self.assertEqual('foo1', nonint2.id)

    def test_creation(self):
        nonint1 = NonIntegerPkFactory.create()
        self.assertEqual('foo0', nonint1.id)

        NonIntegerPkFactory.reset_sequence()
        nonint2 = NonIntegerPkFactory.build()
        self.assertEqual('foo0', nonint2.id)

    def test_force_pk(self):
        nonint1 = NonIntegerPkFactory.create(id='foo10')
        self.assertEqual('foo10', nonint1.id)

        NonIntegerPkFactory.reset_sequence()
        nonint2 = NonIntegerPkFactory.create()
        self.assertEqual('foo0', nonint2.id)


class SQLAlchemyNoSessionTestCase(TransactionTestCase):

    def test_create_raises_exception_when_no_session_was_set(self):
        with self.assertRaises(RuntimeError):
            NoSessionFactory.create()

    def test_build_does_not_raises_exception_when_no_session_was_set(self):
        NoSessionFactory.reset_sequence()  # Make sure we start at test ID 0
        inst0 = NoSessionFactory.build()
        inst1 = NoSessionFactory.build()
        self.assertEqual(inst0.id, 0)
        self.assertEqual(inst1.id, 1)


class SQLAlchemySessionFactoryTestCase(TransactionTestCase):
    def test_create_get_session_from_sqlalchemy_session_factory(self):
        class SessionGetterFactory(SQLAlchemyModelFactory):
            class Meta:
                model = models.StandardModel
                sqlalchemy_session_factory = lambda: models.session

            id = factory.Sequence(lambda n: n)

        SessionGetterFactory.create()
        self.assertEqual(SessionGetterFactory._meta.sqlalchemy_session, models.session)
        # Reuse the session obtained from sqlalchemy_session_factory.
        SessionGetterFactory.create()

    def test_create_raise_exception_sqlalchemy_session_factory_not_callable(self):
        message = "^Provide either a sqlalchemy_session or a sqlalchemy_session_factory, not both$"
        with self.assertRaisesRegex(RuntimeError, message):
            class SessionAndGetterFactory(SQLAlchemyModelFactory):
                class Meta:
                    model = models.StandardModel
                    sqlalchemy_session = models.session
                    sqlalchemy_session_factory = lambda: models.session

                id = factory.Sequence(lambda n: n)


class NameConflictTests(TransactionTestCase):
    """Regression test for `TypeError: _save() got multiple values for argument 'session'`

    See #775.
    """
    def test_no_name_conflict_on_save(self):
        class SpecialFieldWithSaveFactory(SQLAlchemyModelFactory):
            class Meta:
                model = models.SpecialFieldModel
                sqlalchemy_session = models.session

            id = factory.Sequence(lambda n: n)
            session = ''

        saved_child = SpecialFieldWithSaveFactory()
        self.assertEqual(saved_child.session, "")

    def test_no_name_conflict_on_get_or_create(self):
        class SpecialFieldWithGetOrCreateFactory(SQLAlchemyModelFactory):
            class Meta:
                model = models.SpecialFieldModel
                sqlalchemy_get_or_create = ('session',)
                sqlalchemy_session = models.session

            id = factory.Sequence(lambda n: n)
            session = ''

        get_or_created_child = SpecialFieldWithGetOrCreateFactory()
        self.assertEqual(get_or_created_child.session, "")



import demoapp

import factory.alchemy
import factory.fuzzy


class BaseFactory(factory.alchemy.SQLAlchemyModelFactory):
    class Meta:
        abstract = True
        sqlalchemy_session = demoapp.db.session


class UserFactory(BaseFactory):
    class Meta:
        model = demoapp.User

    username = factory.fuzzy.FuzzyText()
    email = factory.fuzzy.FuzzyText()


class UserLogFactory(BaseFactory):
    class Meta:
        model = demoapp.UserLog

    message = factory.fuzzy.FuzzyText()
    user = factory.SubFactory(UserFactory)



# Copyright: See the LICENSE file.

import collections
import datetime
import random
import unittest

import faker.providers

import factory


class MockFaker:
    def __init__(self, expected):
        self.expected = expected
        self.random = random.Random()

    def format(self, provider, **kwargs):
        return self.expected[provider]


class AdvancedMockFaker:
    def __init__(self, handlers):
        self.handlers = handlers
        self.random = random.Random()

    def format(self, provider, **kwargs):
        handler = self.handlers[provider]
        return handler(**kwargs)


class FakerTests(unittest.TestCase):
    def setUp(self):
        self._real_fakers = factory.Faker._FAKER_REGISTRY
        factory.Faker._FAKER_REGISTRY = {}

    def tearDown(self):
        factory.Faker._FAKER_REGISTRY = self._real_fakers

    def _setup_mock_faker(self, locale=None, **definitions):
        if locale is None:
            locale = factory.Faker._DEFAULT_LOCALE
        factory.Faker._FAKER_REGISTRY[locale] = MockFaker(definitions)

    def _setup_advanced_mock_faker(self, locale=None, **handlers):
        if locale is None:
            locale = factory.Faker._DEFAULT_LOCALE
        factory.Faker._FAKER_REGISTRY[locale] = AdvancedMockFaker(handlers)

    def test_simple_biased(self):
        self._setup_mock_faker(name="John Doe")
        faker_field = factory.Faker('name')
        self.assertEqual("John Doe", faker_field.evaluate(None, None, {'locale': None}))

    def test_full_factory(self):
        class Profile:
            def __init__(self, first_name, last_name, email):
                self.first_name = first_name
                self.last_name = last_name
                self.email = email

        class ProfileFactory(factory.Factory):
            class Meta:
                model = Profile
            first_name = factory.Faker('first_name')
            last_name = factory.Faker('last_name', locale='fr_FR')
            email = factory.Faker('email')

        self._setup_mock_faker(first_name="John", last_name="Doe", email="john.doe@example.org")
        self._setup_mock_faker(first_name="Jean", last_name="Valjean", email="jvaljean@exemple.fr", locale='fr_FR')

        profile = ProfileFactory()
        self.assertEqual("John", profile.first_name)
        self.assertEqual("Valjean", profile.last_name)
        self.assertEqual('john.doe@example.org', profile.email)

    def test_override_locale(self):
        class Profile:
            def __init__(self, first_name, last_name):
                self.first_name = first_name
                self.last_name = last_name

        class ProfileFactory(factory.Factory):
            class Meta:
                model = Profile

            first_name = factory.Faker('first_name')
            last_name = factory.Faker('last_name', locale='fr_FR')

        self._setup_mock_faker(first_name="John", last_name="Doe")
        self._setup_mock_faker(first_name="Jean", last_name="Valjean", locale='fr_FR')
        self._setup_mock_faker(first_name="Johannes", last_name="Brahms", locale='de_DE')

        profile = ProfileFactory()
        self.assertEqual("John", profile.first_name)
        self.assertEqual("Valjean", profile.last_name)

        with factory.Faker.override_default_locale('de_DE'):
            profile = ProfileFactory()
            self.assertEqual("Johannes", profile.first_name)
            self.assertEqual("Valjean", profile.last_name)

        profile = ProfileFactory()
        self.assertEqual("John", profile.first_name)
        self.assertEqual("Valjean", profile.last_name)

    def test_add_provider(self):
        class Face:
            def __init__(self, smiley, french_smiley):
                self.smiley = smiley
                self.french_smiley = french_smiley

        class FaceFactory(factory.Factory):
            class Meta:
                model = Face

            smiley = factory.Faker('smiley')
            french_smiley = factory.Faker('smiley', locale='fr_FR')

        class SmileyProvider(faker.providers.BaseProvider):
            def smiley(self):
                return ':)'

        class FrenchSmileyProvider(faker.providers.BaseProvider):
            def smiley(self):
                return '(:'

        factory.Faker.add_provider(SmileyProvider)
        factory.Faker.add_provider(FrenchSmileyProvider, 'fr_FR')

        face = FaceFactory()
        self.assertEqual(":)", face.smiley)
        self.assertEqual("(:", face.french_smiley)

    def test_faker_customization(self):
        """Factory declarations in Faker parameters should be accepted."""
        Trip = collections.namedtuple('Trip', ['departure', 'transfer', 'arrival'])

        may_4th = datetime.date(1977, 5, 4)
        may_25th = datetime.date(1977, 5, 25)
        october_19th = datetime.date(1977, 10, 19)

        class TripFactory(factory.Factory):
            class Meta:
                model = Trip

            departure = may_4th
            arrival = may_25th
            transfer = factory.Faker(
                'date_between_dates',
                start_date=factory.SelfAttribute('..departure'),
                end_date=factory.SelfAttribute('..arrival'),
            )

        def fake_select_date(start_date, end_date):
            """Fake date_between_dates."""
            # Ensure that dates have been transferred from the factory
            # to Faker parameters.
            self.assertEqual(start_date, may_4th)
            self.assertEqual(end_date, may_25th)
            return october_19th

        self._setup_advanced_mock_faker(
            date_between_dates=fake_select_date,
        )

        trip = TripFactory()
        self.assertEqual(may_4th, trip.departure)
        self.assertEqual(october_19th, trip.transfer)
        self.assertEqual(may_25th, trip.arrival)



# Copyright: See the LICENSE file.


"""Helpers for testing SQLAlchemy apps."""

from sqlalchemy import Column, Integer, Unicode, create_engine
from sqlalchemy.orm import declarative_base, scoped_session, sessionmaker

engine_name = 'sqlite://'

session = scoped_session(sessionmaker())
engine = create_engine(engine_name)
session.configure(bind=engine)
Base = declarative_base()


class StandardModel(Base):
    __tablename__ = 'StandardModelTable'

    id = Column(Integer(), primary_key=True)
    foo = Column(Unicode(20))


class MultiFieldModel(Base):
    __tablename__ = 'MultiFieldModelTable'

    id = Column(Integer(), primary_key=True)
    foo = Column(Unicode(20))
    slug = Column(Unicode(20), unique=True)


class MultifieldUniqueModel(Base):
    __tablename__ = 'MultiFieldUniqueModelTable'

    id = Column(Integer(), primary_key=True)
    slug = Column(Unicode(20), unique=True)
    text = Column(Unicode(20), unique=True)
    title = Column(Unicode(20), unique=True)


class NonIntegerPk(Base):
    __tablename__ = 'NonIntegerPk'

    id = Column(Unicode(20), primary_key=True)


class SpecialFieldModel(Base):
    __tablename__ = 'SpecialFieldModelTable'

    id = Column(Integer(), primary_key=True)
    session = Column(Unicode(20))



# Copyright: See the LICENSE file.

import pathlib
import unittest

import factory

SETUP_CFG_VERSION_PREFIX = "version ="


class VersionTestCase(unittest.TestCase):
    def get_setupcfg_version(self):
        setup_cfg_path = pathlib.Path(__file__).parent.parent / "setup.cfg"
        with setup_cfg_path.open("r") as f:
            for line in f:
                if line.startswith(SETUP_CFG_VERSION_PREFIX):
                    return line[len(SETUP_CFG_VERSION_PREFIX):].strip()

    def test_version(self):
        self.assertEqual(factory.__version__, self.get_setupcfg_version())



# Copyright: See the LICENSE file.

"""Helper to test circular factory dependencies."""

import factory

from . import bar as bar_mod


class Foo:
    def __init__(self, bar, x):
        self.bar = bar
        self.x = x


class FooFactory(factory.Factory):
    class Meta:
        model = Foo

    x = 42
    bar = factory.SubFactory(bar_mod.BarFactory)



# Copyright: See the LICENSE file.

from unittest import TestCase

import factory


class TransformCounter:
    calls_count = 0

    @classmethod
    def __call__(cls, x):
        cls.calls_count += 1
        return x.upper()

    @classmethod
    def reset(cls):
        cls.calls_count = 0


transform = TransformCounter()


class Upper:
    def __init__(self, name, **extra):
        self.name = name
        self.extra = extra


class UpperFactory(factory.Factory):
    name = factory.Transformer("value", transform=transform)

    class Meta:
        model = Upper


class TransformerTest(TestCase):
    def setUp(self):
        transform.reset()

    def test_transform_count(self):
        self.assertEqual("VALUE", UpperFactory().name)
        self.assertEqual(transform.calls_count, 1)

    def test_transform_kwarg(self):
        self.assertEqual("TEST", UpperFactory(name="test").name)
        self.assertEqual(transform.calls_count, 1)
        self.assertEqual("VALUE", UpperFactory().name)
        self.assertEqual(transform.calls_count, 2)

    def test_transform_faker(self):
        value = UpperFactory(name=factory.Faker("first_name_female", locale="fr")).name
        self.assertIs(value.isupper(), True)

    def test_transform_linked(self):
        value = UpperFactory(
            name=factory.LazyAttribute(lambda o: o.username.replace(".", " ")),
            username="john.doe",
        ).name
        self.assertEqual(value, "JOHN DOE")

    def test_force_value(self):
        value = UpperFactory(name=factory.Transformer.Force("Mia")).name
        self.assertEqual(value, "Mia")

    def test_force_value_declaration(self):
        """Pretty unlikely use case, but easy enough to cover."""
        value = UpperFactory(
            name=factory.Transformer.Force(
                factory.LazyFunction(lambda: "infinity")
            )
        ).name
        self.assertEqual(value, "infinity")

    def test_force_value_declaration_context(self):
        """Ensure "forced" values run at the right level."""
        value = UpperFactory(
            name=factory.Transformer.Force(
                factory.LazyAttribute(lambda o: o.username.replace(".", " ")),
            ),
            username="john.doe",
        ).name
        self.assertEqual(value, "john doe")


class TestObject:
    def __init__(self, one=None, two=None, three=None):
        self.one = one
        self.two = two
        self.three = three


class TransformDeclarationFactory(factory.Factory):
    class Meta:
        model = TestObject
    one = factory.Transformer("", transform=str.upper)
    two = factory.Transformer(factory.Sequence(int), transform=lambda n: n ** 2)


class TransformerSequenceTest(TestCase):
    def test_on_sequence(self):
        instance = TransformDeclarationFactory(__sequence=2)
        self.assertEqual(instance.one, "")
        self.assertEqual(instance.two, 4)
        self.assertIsNone(instance.three)

    def test_on_user_supplied(self):
        """A transformer can wrap a call-time declaration"""
        instance = TransformDeclarationFactory(
            one=factory.Sequence(str),
            two=2,
            __sequence=2,
        )
        self.assertEqual(instance.one, "2")
        self.assertEqual(instance.two, 4)
        self.assertIsNone(instance.three)


class WithMaybeFactory(factory.Factory):
    class Meta:
        model = TestObject

    one = True
    two = factory.Maybe(
        'one',
        yes_declaration=factory.Transformer("yes", transform=str.upper),
        no_declaration=factory.Transformer("no", transform=str.upper),
    )
    three = factory.Maybe('one', no_declaration=factory.Transformer("three", transform=str.upper))


class TransformerMaybeTest(TestCase):
    def test_default_transform(self):
        instance = WithMaybeFactory()
        self.assertIs(instance.one, True)
        self.assertEqual(instance.two, "YES")
        self.assertIsNone(instance.three)

    def test_yes_transform(self):
        instance = WithMaybeFactory(one=True)
        self.assertIs(instance.one, True)
        self.assertEqual(instance.two, "YES")
        self.assertIsNone(instance.three)

    def test_no_transform(self):
        instance = WithMaybeFactory(one=False)
        self.assertIs(instance.one, False)
        self.assertEqual(instance.two, "NO")
        self.assertEqual(instance.three, "THREE")

    def test_override(self):
        instance = WithMaybeFactory(one=True, two="NI")
        self.assertIs(instance.one, True)
        self.assertEqual(instance.two, "NI")
        self.assertIsNone(instance.three)


class RelatedTest(TestCase):
    def test_default_transform(self):
        cities = []

        class City:
            def __init__(self, capital_of, name):
                self.capital_of = capital_of
                self.name = name
                cities.append(self)

        class Country:
            def __init__(self, name):
                self.name = name

        class CityFactory(factory.Factory):
            class Meta:
                model = City

            name = "Rennes"

        class CountryFactory(factory.Factory):
            class Meta:
                model = Country

            name = "France"
            capital_city = factory.RelatedFactory(
                CityFactory,
                factory_related_name="capital_of",
                name=factory.Transformer("Paris", transform=str.upper),
            )

        instance = CountryFactory()
        self.assertEqual(instance.name, "France")
        [city] = cities
        self.assertEqual(city.capital_of, instance)
        self.assertEqual(city.name, "PARIS")


class WithTraitFactory(factory.Factory):
    class Meta:
        model = TestObject

    class Params:
        upper_two = factory.Trait(
            two=factory.Transformer("two", transform=str.upper)
        )
        odds = factory.Trait(
            one="one",
            three="three",
        )
    one = factory.Transformer("one", transform=str.upper)


class TransformerTraitTest(TestCase):
    def test_traits_off(self):
        instance = WithTraitFactory()
        self.assertEqual(instance.one, "ONE")
        self.assertIsNone(instance.two)
        self.assertIsNone(instance.three)

    def test_trait_transform_applies(self):
        """A trait-provided transformer should apply to existing values"""
        instance = WithTraitFactory(upper_two=True)
        self.assertEqual(instance.one, "ONE")
        self.assertEqual(instance.two, "TWO")
        self.assertIsNone(instance.three)

    def test_trait_transform_applies_supplied(self):
        """A trait-provided transformer should be overridden by caller-provided values"""
        instance = WithTraitFactory(upper_two=True, two="two")
        self.assertEqual(instance.one, "ONE")
        self.assertEqual(instance.two, "two")
        self.assertIsNone(instance.three)



# Copyright: See the LICENSE file.


"""factory_boy extensions for use with the Django framework."""


import functools
import io
import logging
import os
import warnings
from typing import Dict, TypeVar

from django.contrib.auth.hashers import make_password
from django.core import files as django_files
from django.db import IntegrityError

from . import base, declarations, errors

logger = logging.getLogger('factory.generate')


DEFAULT_DB_ALIAS = 'default'  # Same as django.db.DEFAULT_DB_ALIAS
T = TypeVar("T")

_LAZY_LOADS: Dict[str, object] = {}


def get_model(app, model):
    """Wrapper around django's get_model."""
    if 'get_model' not in _LAZY_LOADS:
        _lazy_load_get_model()

    _get_model = _LAZY_LOADS['get_model']
    return _get_model(app, model)


def _lazy_load_get_model():
    """Lazy loading of get_model.

    get_model loads django.conf.settings, which may fail if
    the settings haven't been configured yet.
    """
    from django import apps as django_apps
    _LAZY_LOADS['get_model'] = django_apps.apps.get_model


class DjangoOptions(base.FactoryOptions):
    def _build_default_options(self):
        return super()._build_default_options() + [
            base.OptionDefault('django_get_or_create', (), inherit=True),
            base.OptionDefault('database', DEFAULT_DB_ALIAS, inherit=True),
            base.OptionDefault('skip_postgeneration_save', False, inherit=True),
        ]

    def _get_counter_reference(self):
        counter_reference = super()._get_counter_reference()
        if (counter_reference == self.base_factory
                and self.base_factory._meta.model is not None
                and self.base_factory._meta.model._meta.abstract
                and self.model is not None
                and not self.model._meta.abstract):
            # Target factory is for an abstract model, yet we're for another,
            # concrete subclass => don't reuse the counter.
            return self.factory
        return counter_reference

    def get_model_class(self):
        if isinstance(self.model, str) and '.' in self.model:
            app, model_name = self.model.split('.', 1)
            self.model = get_model(app, model_name)

        return self.model


class DjangoModelFactory(base.Factory[T]):
    """Factory for Django models.

    This makes sure that the 'sequence' field of created objects is a new id.

    Possible improvement: define a new 'attribute' type, AutoField, which would
    handle those for non-numerical primary keys.
    """

    _options_class = DjangoOptions
    _original_params = None

    class Meta:
        abstract = True  # Optional, but explicit.

    @classmethod
    def _load_model_class(cls, definition):

        if isinstance(definition, str) and '.' in definition:
            app, model = definition.split('.', 1)
            return get_model(app, model)

        return definition

    @classmethod
    def _get_manager(cls, model_class):
        if model_class is None:
            raise errors.AssociatedClassError(
                f"No model set on {cls.__module__}.{cls.__name__}.Meta")

        try:
            manager = model_class.objects
        except AttributeError:
            # When inheriting from an abstract model with a custom
            # manager, the class has no 'objects' field.
            manager = model_class._default_manager

        if cls._meta.database != DEFAULT_DB_ALIAS:
            manager = manager.using(cls._meta.database)
        return manager

    @classmethod
    def _generate(cls, strategy, params):
        # Original params are used in _get_or_create if it cannot build an
        # object initially due to an IntegrityError being raised
        cls._original_params = params
        return super()._generate(strategy, params)

    @classmethod
    def _get_or_create(cls, model_class, *args, **kwargs):
        """Create an instance of the model through objects.get_or_create."""
        manager = cls._get_manager(model_class)

        assert 'defaults' not in cls._meta.django_get_or_create, (
            "'defaults' is a reserved keyword for get_or_create "
            "(in %s._meta.django_get_or_create=%r)"
            % (cls, cls._meta.django_get_or_create))

        key_fields = {}
        for field in cls._meta.django_get_or_create:
            if field not in kwargs:
                raise errors.FactoryError(
                    "django_get_or_create - "
                    "Unable to find initialization value for '%s' in factory %s" %
                    (field, cls.__name__))
            key_fields[field] = kwargs.pop(field)
        key_fields['defaults'] = kwargs

        try:
            instance, _created = manager.get_or_create(*args, **key_fields)
        except IntegrityError as e:

            if cls._original_params is None:
                raise e

            get_or_create_params = {
                lookup: value
                for lookup, value in cls._original_params.items()
                if lookup in cls._meta.django_get_or_create
            }
            if get_or_create_params:
                try:
                    instance = manager.get(**get_or_create_params)
                except manager.model.DoesNotExist:
                    # Original params are not a valid lookup and triggered a create(),
                    # that resulted in an IntegrityError. Follow Django’s behavior.
                    raise e
            else:
                raise e

        return instance

    @classmethod
    def _create(cls, model_class, *args, **kwargs):
        """Create an instance of the model, and save it to the database."""
        if cls._meta.django_get_or_create:
            return cls._get_or_create(model_class, *args, **kwargs)

        manager = cls._get_manager(model_class)
        return manager.create(*args, **kwargs)

    # DEPRECATED. Remove this override with the next major release.
    @classmethod
    def _after_postgeneration(cls, instance, create, results=None):
        """Save again the instance if creating and at least one hook ran."""
        if create and results and not cls._meta.skip_postgeneration_save:
            warnings.warn(
                f"{cls.__name__}._after_postgeneration will stop saving the instance "
                "after postgeneration hooks in the next major release.\n"
                "If the save call is extraneous, set skip_postgeneration_save=True "
                f"in the {cls.__name__}.Meta.\n"
                "To keep saving the instance, move the save call to your "
                "postgeneration hooks or override _after_postgeneration.",
                DeprecationWarning,
            )
            # Some post-generation hooks ran, and may have modified us.
            instance.save()


class Password(declarations.Transformer):
    def __init__(self, password, transform=make_password, **kwargs):
        super().__init__(password, transform=transform, **kwargs)


class FileField(declarations.BaseDeclaration):
    """Helper to fill in django.db.models.FileField from a Factory."""

    DEFAULT_FILENAME = 'example.dat'

    def _make_data(self, params):
        """Create data for the field."""
        return params.get('data', b'')

    def _make_content(self, params):
        path = ''

        from_path = params.get('from_path')
        from_file = params.get('from_file')
        from_func = params.get('from_func')

        if len([p for p in (from_path, from_file, from_func) if p]) > 1:
            raise ValueError(
                "At most one argument from 'from_file', 'from_path', and 'from_func' should "
                "be non-empty when calling factory.django.FileField."
            )

        if from_path:
            path = from_path
            with open(path, 'rb') as f:
                content = django_files.base.ContentFile(f.read())

        elif from_file:
            f = from_file
            content = django_files.File(f)
            path = content.name

        elif from_func:
            func = from_func
            content = django_files.File(func())
            path = content.name

        else:
            data = self._make_data(params)
            content = django_files.base.ContentFile(data)

        if path:
            default_filename = os.path.basename(path)
        else:
            default_filename = self.DEFAULT_FILENAME

        filename = params.get('filename', default_filename)
        return filename, content

    def evaluate(self, instance, step, extra):
        """Fill in the field."""
        filename, content = self._make_content(extra)
        return django_files.File(content.file, filename)


class ImageField(FileField):
    DEFAULT_FILENAME = 'example.jpg'

    def _make_data(self, params):
        # ImageField (both django's and factory_boy's) require PIL.
        # Try to import it along one of its known installation paths.
        from PIL import Image

        width = params.get('width', 100)
        height = params.get('height', width)
        color = params.get('color', 'blue')
        image_format = params.get('format', 'JPEG')
        image_palette = params.get('palette', 'RGB')

        thumb_io = io.BytesIO()
        with Image.new(image_palette, (width, height), color) as thumb:
            thumb.save(thumb_io, format=image_format)
        return thumb_io.getvalue()


class mute_signals:
    """Temporarily disables and then restores any django signals.

    Args:
        *signals (django.dispatch.dispatcher.Signal): any django signals

    Examples:
        with mute_signals(pre_init):
            user = UserFactory.build()
            ...

        @mute_signals(pre_save, post_save)
        class UserFactory(factory.Factory):
            ...

        @mute_signals(post_save)
        def generate_users():
            UserFactory.create_batch(10)
    """

    def __init__(self, *signals):
        self.signals = signals
        self.paused = {}

    def __enter__(self):
        for signal in self.signals:
            logger.debug('mute_signals: Disabling signal handlers %r',
                         signal.receivers)

            # Note that we're using implementation details of
            # django.signals, since arguments to signal.connect()
            # are lost in signal.receivers
            self.paused[signal] = signal.receivers
            signal.receivers = []

    def __exit__(self, exc_type, exc_value, traceback):
        for signal, receivers in self.paused.items():
            logger.debug('mute_signals: Restoring signal handlers %r',
                         receivers)

            signal.receivers = receivers + signal.receivers
            with signal.lock:
                # Django uses some caching for its signals.
                # Since we're bypassing signal.connect and signal.disconnect,
                # we have to keep messing with django's internals.
                signal.sender_receivers_cache.clear()
        self.paused = {}

    def copy(self):
        return mute_signals(*self.signals)

    def __call__(self, callable_obj):
        if isinstance(callable_obj, base.FactoryMetaClass):
            # Retrieve __func__, the *actual* callable object.
            callable_obj._create = self.wrap_method(callable_obj._create.__func__)
            callable_obj._generate = self.wrap_method(callable_obj._generate.__func__)
            callable_obj._after_postgeneration = self.wrap_method(
                callable_obj._after_postgeneration.__func__
            )
            return callable_obj

        else:
            @functools.wraps(callable_obj)
            def wrapper(*args, **kwargs):
                # A mute_signals() object is not reentrant; use a copy every time.
                with self.copy():
                    return callable_obj(*args, **kwargs)
            return wrapper

    def wrap_method(self, method):
        @classmethod
        @functools.wraps(method)
        def wrapped_method(*args, **kwargs):
            # A mute_signals() object is not reentrant; use a copy every time.
            with self.copy():
                return method(*args, **kwargs)
        return wrapped_method
